-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM
--
-- Restricted Materials of IBM
--
-- 5724-C51
--
-- (C) Copyright IBM Corp. 2004 All Rights Reserved.
--
-- US Government Users Restricted Rights - Use, duplication or
-- disclosure restricted by GSA ADP Schedule Contract with
-- IBM Corp.
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/asisp_processstagednotes.sql
-- SID: 1.4
-- Date Modified: 05/01/04 16:48:15
--
-- Module Description
--   Defines the asisp_processStagedNotes stored procedure
USE Object
GO

-- asisp_processStagedNotes
PRINT 'Creating procedure asisp_processStagedNotes'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'asisp_processStagedNotes' AND type = 'P')
  DROP PROCEDURE asisp_processStagedNotes
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM
--
-- Restricted Materials of IBM
--
-- 5724-C51
--
-- (C) Copyright IBM Corp. 2004 All Rights Reserved.
--
-- US Government Users Restricted Rights - Use, duplication or
-- disclosure restricted by GSA ADP Schedule Contract with
-- IBM Corp.
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/asisp_processstagednotes.sql
-- SID: 1.4
-- Date Modified: 05/01/04 16:48:15
--
-- Processes ownership notes.
-- When an ownership note is created a record is inserted into the StagedNote table after the
-- note's OwnedEvent records have been inserted. This procedure performs the remainder of the
-- processing for these notes, which is primarly the following:
-- 1. Create the NoteDependent records
-- 2. Negate owned events for notes configured to clear events
-- 3. Update OpenEventCount for notes not configured to clear events
CREATE PROCEDURE asisp_processStagedNotes
  @maxrows  INT = 0, -- ignored
  @verbose  TINYINT = 0
AS
  SET NOCOUNT ON

  -- table to hold StagedNote records being processed
  DECLARE @note TABLE
  (
    timestamp     BINARY(8)     NOT NULL,
    id            INT           NOT NULL,
    FinalStatus   NVARCHAR(16)  NOT NULL,
    Flags         INT           NOT NULL,
    Status        NVARCHAR(16)  NOT NULL,
    CreatorCID    NVARCHAR(4)   NOT NULL,
    CreatorObjID  INT           NOT NULL,
    UserID        NVARCHAR(255) NOT NULL
  )
  
  -- table to hold new NoteDependent records
  DECLARE @nd TABLE
  (
    noteid        INT           NOT NULL,
    origcid       NVARCHAR(4)   NOT NULL,
    origid        INT           NOT NULL,
    depcid        NVARCHAR(4)   NOT NULL,
    depid         INT           NOT NULL,
    eventcid      NVARCHAR(4)   NOT NULL,
    eventid       INT           NOT NULL,
    lob_only      TINYINT       NOT NULL,
    status        TINYINT       NOT NULL	-- 0 (open), 1 (cleared), 2 (expired), 3 (closed)
  )

  -- table for set of updates to OpenEventCount nevents for parents of the exceptions
  DECLARE @upd TABLE
  (
    cid           NVARCHAR(4)   NOT NULL,
    id            INT           NOT NULL,
    nevents       INT           NOT NULL
  )

  -- table for recording for all objects that are becoming NoteDependents due to red exceptions or
  -- messages
  DECLARE @hist TABLE
  (
    cid           NVARCHAR(4)   NOT NULL,
    id            INT           NOT NULL
  )

  -- table for updates to HasOwnershipNote
  -- This must be a temp table instead of table variable becuase it is used in dynamic SQL
  CREATE TABLE #dep
  (
    depcid        ClassID       NOT NULL,
    depid         ObjID         NOT NULL
  )

  -- Variable declarations
  DECLARE
    @rc               INT,
    @err              INT,
    @now              NVARCHAR(40),
    @tcount           INT,
    @TEC_NOTE_SUBJECT NVARCHAR(255),
    @notify_cid       ClassID,
    @sql              NVARCHAR(1000),
    @cname            NVARCHAR(24)

  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Starting', 0, 1, @now, @rc) WITH NOWAIT
  END
  
  INSERT INTO @note
    SELECT s.timestamp, s.id, s.FinalStatus, n._Flags, n._Status, n._CreatorCID, n._CreatorObjID, n._UserID
    FROM StagedNote AS s
    JOIN Note_A AS n WITH (NOLOCK) ON n.id = s.id
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: error %d selecting from StagedNote', 11, 1, @now, @err) WITH NOWAIT
    GOTO ErrorExit
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Found %d staged note(s)', 0, 1, @now, @rc) WITH NOWAIT
  END
  
  --
  -- Create the NoteDependent records
  --

  -- The objects that events belong to become NoteDependents. These are the "direct" dependents.
  INSERT INTO @nd(noteid, origcid, origid, depcid, depid, eventcid, eventid, lob_only, status)
    SELECT n.id, o.parentcid, o.parentid, o.parentcid, o.parentid, o.eventcid, o.eventid, 0, o.status
    FROM OwnedEvent AS o
      JOIN @note AS n ON n.id = o.noteid
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: error %d computing direct dependents', 11, 1, @now, @err) WITH NOWAIT
    GOTO ErrorExit
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Found %d direct dependent(s)', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- If the object that owns the events is not also the object from which the note is created, then
  -- the latter object must also be made a note dependent
  IF EXISTS
  (
    SELECT 1
    FROM @nd AS nd
    JOIN @note AS n ON n.id = nd.noteid
    WHERE nd.origcid <> n.CreatorCID
       OR nd.origid  <> n.CreatorObjID
       OR nd.depcid  <> n.CreatorCID
       OR nd.depid   <> n.CreatorObjID
  )
  BEGIN
    INSERT INTO @nd(noteid, origcid, origid, depcid, depid, eventcid, eventid, lob_only, status)
      SELECT nd.noteid, n.CreatorCID, n.CreatorObjID, n.CreatorCID, n.CreatorObjID, nd.eventcid, nd.eventid, 0, nd.status
      FROM @nd AS nd
        JOIN @note AS n ON n.id = nd.noteid
        LEFT OUTER JOIN @nd AS orig ON orig.noteid  = nd.noteid
                                  AND orig.origcid = n.CreatorCID
                                  AND orig.origid  = n.CreatorObjID
                                  AND orig.depcid  = n.CreatorCID
                                  AND orig.depid   = n.CreatorObjID
      WHERE orig.noteid IS NULL
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: error %d computing indirect dependents', 11, 1, @now, @err) WITH NOWAIT
      GOTO ErrorExit
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Found %d indirect dependent(s)', 0, 1, @now, @rc) WITH NOWAIT
    END
  END
  
  -- Normally, any LOBs linked to the objects to which the events belong also become note dependents,
  -- unless the events have the LOB-Only flag set, so we first see if there are any linked LOBs.
  INSERT INTO @nd(noteid, origcid, origid, depcid, depid, eventcid, eventid, lob_only, status)
    SELECT n.noteid, n.depcid, n.depid, 'LOB', ll.dst_id, n.eventcid, n.eventid, 0, n.status
    FROM @nd AS n
    JOIN lob_link AS ll ON ll.phy_cid = n.depcid AND ll.phy_id = n.depid
    WHERE ll.src_id = 0
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: error %d computing set of linked dependents', 11, 1, @now, @err) WITH NOWAIT
    GOTO ErrorExit
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Found %d linked dependent(s)', 0, 1, @now, @rc) WITH NOWAIT
  END
  IF @rc > 0
  BEGIN -- Remove records for LOB-only events where the dependent is linked to the parent
    UPDATE n SET lob_only = 1
    FROM @nd AS n
    JOIN Exception_V AS e WITH (NOLOCK) ON e.id = n.eventid
    WHERE n.eventcid = 'EXCP' AND e._flags & 16 = 16
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: error %d looking up LOB-only exceptions', 11, 1, @now, @err) WITH NOWAIT
      GOTO ErrorExit
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Found %d LOB-only exception(s)', 0, 1, @now, @rc) WITH NOWAIT
    END

    UPDATE n SET lob_only = 1
    FROM @nd AS n
    JOIN Message_V AS e WITH (NOLOCK) ON e.id = n.eventid
    WHERE n.eventcid = 'MESG' AND e._flags & 16 = 16
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: error %d looking up LOB-only messages', 11, 1, @now, @err) WITH NOWAIT
      GOTO ErrorExit
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Found %d LOB-only message(s)', 0, 1, @now, @rc) WITH NOWAIT
    END

    IF EXISTS (SELECT 1 FROM @nd WHERE lob_only = 1)
    BEGIN
      DELETE n FROM @nd AS n
      JOIN lob_link AS l ON l.dst_id = n.depid AND l.phy_cid = n.origcid AND l.phy_id = n.origid
      WHERE n.depcid = 'LOB' AND n.lob_only = 1
      SELECT @err = @@ERROR, @rc = @@ROWCOUNT
      IF @err <> 0
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: asisp_processStagedNotes: error %d removing LOB-only NoteDependent updates', 11, 1, @now, @err) WITH NOWAIT
        GOTO ErrorExit
      END
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: asisp_processStagedNotes: Removed %d LOB-only NoteDependent update(s)', 0, 1, @now, @rc) WITH NOWAIT
      END
    END
  END -- Remove records for LOB-only events where the dependent is linked to the parent

  IF @verbose >= 2
  BEGIN
    RAISERROR('%s: asisp_processStagedNotes: dumping temp table @nd', 11, 1, @now) WITH NOWAIT
    SELECT '@nd' AS '@nd', * FROM @nd
  END

  -- Lookup objects that are becoming dependents of notes with red messages
  INSERT INTO @hist(cid, id)
    SELECT DISTINCT i.depcid, i.depid
    FROM @nd AS i
    JOIN Message_V AS e WITH (NOLOCK) ON i.eventid = e.id
    WHERE e._AlertState = 3
      AND i.eventcid = 'MESG'
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Failed looking up messages. Error=%d', 11, 1, @now, @err) WITH NOWAIT
    GOTO ErrorExit
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Found %d red message(s)', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- Lookup objects that are becoming dependents of notes with red exceptions
  INSERT INTO @hist(cid, id)
    SELECT DISTINCT i.depcid, i.depid
    FROM @nd AS i
    JOIN Exception_V AS e WITH (NOLOCK) ON i.eventid = e.id
      LEFT OUTER JOIN @hist AS hprev ON hprev.cid = i.depcid AND hprev.id = i.depid
    WHERE e._AlertState = 3
      AND i.eventcid = 'EXCP'
      AND hprev.cid IS NULL
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Failed looking up exceptions. Error=%d', 11, 1, @now, @err) WITH NOWAIT
    GOTO ErrorExit
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Found %d red exception(s)', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- Remove from the @hist table any objects that already have NoteDependent records from red messages
  DELETE h
  FROM @hist AS h
  JOIN NoteDependent AS prev WITH (HOLDLOCK) ON prev.depcid = h.cid AND prev.depid = h.id
  JOIN Message_V AS e WITH (NOLOCK) ON e.id = prev.eventid
  WHERE prev.eventcid = 'MESG'
    AND prev.status   = 0
    AND e._AlertState = 3
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Error %d looking up previous NoteDependent records from red messages', 11, 1, @now, @err) WITH NOWAIT
    GOTO ErrorExit
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Removed %d record(s) from ObjectStateHist insert set due to previously owned red messages', 0, 1, @now) WITH NOWAIT
  END

  -- Remove from the @hist table any objects that already have NoteDependent records from red exceptions
  DELETE h
  FROM @hist AS h
  JOIN NoteDependent AS prev WITH (HOLDLOCK) ON prev.depcid = h.cid AND prev.depid = h.id
  JOIN Exception_V AS e WITH (NOLOCK) ON e.id = prev.eventid
  WHERE prev.eventcid = 'EXCP'
    AND prev.status   = 0
    AND e._AlertState = 3
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Error %d looking up previous NoteDependent records from red exceptions', 11, 1, @now, @err) WITH NOWAIT
    GOTO ErrorExit
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Removed %d record(s) from ObjectStateHist insert set due to previously owned red exceptions', 0, 1, @now) WITH NOWAIT
  END

  -- From this point forward we start modifying permanent tables, so we need a transaction
  SET @tcount = @@TRANCOUNT -- don't set this any earlier because ErrorExit relies on this being non-null to trigger the rollback if needed
  IF @tcount = 0
    BEGIN TRANSACTION asisp_processStagedNotes
  ELSE
    SAVE TRANSACTION asisp_processStagedNotes

  -- Get the latest status from OwnedEvent just before inserting into NoteDependent
  UPDATE n SET status = o.status
  FROM @nd AS n
    JOIN OwnedEvent AS o WITH (HOLDLOCK) ON n.eventcid = o.eventcid AND n.eventid = o.eventid
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: error %d updating status in @nd from OwnedEvent', 11, 1, @now, @err) WITH NOWAIT
    GOTO ErrorExit
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Updated status of %d record(s) in @nd from OwnedEvent', 0, 1, @now, @rc) WITH NOWAIT
  END

  INSERT INTO NoteDependent(noteid, depcid, depid, eventcid, eventid, status)
    SELECT DISTINCT noteid, depcid, depid, eventcid, eventid, status
    FROM @nd
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: error %d inserting into table NoteDependent', 11, 1, @now, @err) WITH NOWAIT
    GOTO ErrorExit
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Inserted %d record(s) into NoteDependent', 0, 1, @now, @rc) WITH NOWAIT
  END
  
  IF EXISTS (SELECT 1 FROM @hist)
  BEGIN
    -- insert distinct records
    INSERT INTO ObjectStateHist(Obj_cid, Obj_id, State_flag, Evt_detail, State_Nm, CTIME)
      SELECT cid, id, 1, 'Open ownership note', 'Owned', CURRENT_TIMESTAMP
      FROM @hist
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Failed inserting into ObjectStateHist. Error=%d', 11, 1, @now, @err) WITH NOWAIT
      ROLLBACK TRANSACTION
      RETURN
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Inserted %d row(s) into ObjectStateHist', 0, 1, @now, @rc) WITH NOWAIT
    END
  END
  
  -- Iterate through the classes of note dependents and set the hasownershipnote flag on the
  -- instances of each class that have dependents when status = 0 (open notes)
  INSERT INTO #dep
    SELECT DISTINCT depcid, depid FROM @nd
    WHERE status = 0 
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Error %d computing set of distinct dependents', 11, 1, @now, @err) WITH NOWAIT
    ROLLBACK TRANSACTION
    RETURN
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Found distinct %d dependent(s)', 0, 1, @now, @rc) WITH NOWAIT
  END
  
  DECLARE ownership_c CURSOR FOR
    SELECT DISTINCT d.depcid, o.cname
    FROM #dep AS d
    JOIN Meta..obj_class_table_BASE AS o WITH (NOLOCK) ON o.cid = d.depcid
  OPEN ownership_c
  FETCH NEXT FROM ownership_c INTO @notify_cid, @cname
  WHILE @@FETCH_STATUS <> -1
  BEGIN
    -- Update the _HasOwnershipNote flag
    SET @sql = 
      'UPDATE obj SET _HasOwnershipNote = 1 ' +
      'FROM ' + @cname + '_V AS obj ' +
      'JOIN #dep AS d WITH (NOLOCK) ON d.depcid = obj.cid AND d.depid = obj.id ' +
      'WHERE obj._HasOwnershipNote = 0 '
    EXEC(@sql)
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: error %d executing SQL [%s]', 3, 1, @now, @err, @sql) WITH NOWAIT
      DEALLOCATE ownership_c
      GOTO ErrorExit
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Updated %d row(s) with SQL [%s]', 0, 1, @now, @rc, @sql) WITH NOWAIT
    END

    FETCH NEXT FROM ownership_c INTO @notify_cid, @cname
  END
  DEALLOCATE ownership_c

  -- Generate negating events for the notes with flag bit 1 (clear) set.
  IF EXISTS (SELECT 1 FROM @note WHERE Flags & 1  = 1)
  BEGIN

    INSERT INTO StagedEXCP
    (
      parent_cid,
      parent_id,
      TIME,
      SEQ_NO,
      OBJECT_NAME,
      NATIVE_KEY,
      EXCP_OBJ_TYPE,
      EXCP_OBJ_NAME,
      EXCP_NAME,
      EXCP_CD,
      nEXCH,
      AlertState,
      Priority
    )
    SELECT
      e._ParentCID,
      e._ParentID,
      CURRENT_TIMESTAMP,
      e._SEQ_NO,
      e._OBJECT_NAME,
      e._NATIVE_KEY,
      e._EXCP_OBJ_TYPE,
      e._EXCP_OBJ_NAME,
      e._EXCP_NAME,
      N'SC - Ownership taken by ' + n.UserID + N' at ' + CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121),
      e._nEXCH + 1,
      e._AlertState,
      e._Priority
    FROM
      @note AS n
      JOIN OwnedEvent AS oe WITH (NOLOCK) ON oe.noteid = n.id
      JOIN Exception_V AS e WITH (NOLOCK) ON e.id = oe.eventid
    WHERE n.Flags & 1 = 1
      AND oe.eventcid = 'EXCP'
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: error %d inserting into StagedEXCP', 11, 1, @now, @err) WITH NOWAIT
      GOTO ErrorExit
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Inserted %d record(s) into StagedEXCP', 0, 1, @now, @rc) WITH NOWAIT
    END

    INSERT INTO StagedMESG
    (
      parent_cid,
      parent_id,
      TIME,
      SEQ_NO,
      OBJECT_NAME,
      NATIVE_KEY,
      MESSAGE,
      REASON_CD,
      AlertState,
      Priority
    )
    SELECT
      e._ParentCID,
      e._ParentID,
      CURRENT_TIMESTAMP,
      e._SEQ_NO,
      e._OBJECT_NAME,
      e._NATIVE_KEY,
      'Ownership', -- MESSAGE
      N'Ownership taken by ' + n.UserID + N' at ' + CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121),
      1, -- AlertState
      e._Priority
    FROM
      @note AS n
      JOIN OwnedEvent AS oe WITH (NOLOCK) ON oe.noteid = n.id
      JOIN Message_V AS e WITH (NOLOCK) ON e.id = oe.eventid
    WHERE n.Flags & 1 = 1
      AND oe.eventcid = 'MESG'
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: error %d inserting into StagedMESG', 11, 1, @now, @err) WITH NOWAIT
      GOTO ErrorExit
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Inserted %d record(s) into StagedMESG', 0, 1, @now, @rc) WITH NOWAIT
    END
  END -- Generate negating events for the notes with flag bit 1 (clear) set.
  
  -- For notes with alerts not cleared by takeownership, update the OpenEventCount table
  IF EXISTS (SELECT 1 FROM @note WHERE Flags & 1 = 0)
  BEGIN
    -- Compute set of updates to OpenEventCount nevents for parents of the exceptions

    -- Updates for resources directly owned by events
    INSERT INTO @upd
    SELECT DISTINCT o.parentcid, o.parentid, COUNT(*)
    FROM OwnedEvent AS o
      JOIN @note AS n ON n.id = o.noteid
    WHERE n.Flags & 1 = 0
    GROUP BY o.parentcid, o.parentid
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Error %d computing entries to update OpenEvenCount (when alerts not cleared)', 11, 1, @now, @err) WITH NOWAIT
      GOTO ErrorExit
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Computed %d rows to update OpenEventCount (when alerts not cleared)', 0, 1, @now, @rc) WITH NOWAIT
    END

    -- Updates of resources indirectly owned by events (LOB)
    INSERT INTO @upd
    SELECT 'LOB', nd.depid, u.nevents
    FROM @upd AS u
    JOIN @nd AS nd ON nd.origcid = u.cid AND nd.origid = u.id
    WHERE nd.depcid = 'LOB'
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Error %d computing dependent (LOB) entries to update OpenEvenCount (when alerts not cleared)', 11, 1, @now, @err) WITH NOWAIT
      GOTO ErrorExit
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Computed %d depended (LOB) rows to update OpenEventCount (when alerts not cleared)', 0, 1, @now, @rc) WITH NOWAIT
    END

    UPDATE oec SET nevents = oec.nevents - u.nevents
    FROM OpenEventCount AS oec
    JOIN @upd AS u ON u.cid = oec.cid AND u.id = oec.id
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Error %d updating OpenEvenCount (when alerts not cleared)', 11, 1, @now, @err) WITH NOWAIT
      GOTO ErrorExit
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: updated %d rows in OpenEventCount (when alerts not cleared)', 0, 1, @now, @rc) WITH NOWAIT
    END
  END -- For notes with alerts not cleared by takeownership, update the OpenEventCount table

  -- Insert PendingNotification records
  INSERT INTO PendingNotification (type, fields, detail)
  SELECT DISTINCT
    et.type,
    CAST(oc.cno AS NVARCHAR(9)) + '|' + CAST(nd.depid AS NVARCHAR(9)) + '|' + CAST(sm.id AS NVARCHAR(9)),
    '1'
  FROM
    EventType AS et WITH (NOLOCK),
    id_string_map AS sm WITH (NOLOCK),
    @nd AS nd
    JOIN Meta..obj_class_table_BASE AS oc WITH (NOLOCK) ON oc.cid = nd.depcid
  WHERE et.name   = 'ObjAttrUpdated'
    AND sm.string = 'HasOwnershipNote'
    AND nd.status = 0
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: error %d inserting into PendingNotification', 11, 1, @now, @err) WITH NOWAIT
    GOTO ErrorExit
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Inserted %d record(s) into PendingNotification', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- Log the change in Owned state for where _HasOwnershipNote is set to 1
  INSERT INTO ObjectStateHist(Obj_cid, Obj_id, CTIME, State_flag, Evt_detail, State_Nm) 
    SELECT DISTINCT nd.depcid, nd.depid, CURRENT_TIMESTAMP, 1, 'Entered owned state.', 'Owned' 
    FROM @nd AS nd
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: error %d inserting into ObjectStateHist for noteid %d', 11, 1, @now, @err) WITH NOWAIT
    GOTO ErrorExit
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Inserted %d record(s) into ObjectStateHist', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- Handle notification to TEC
  IF EXISTS -- Is TEC Notification enabled
  (
    SELECT 1
    FROM SystemConfiguration WITH (NOLOCK)
    WHERE groupname='TEC_EVENT_OWNERSHIP'
      AND propertyname='NOTIFY_TEC_OF_TBSM_OWNERSHIP'
      AND UPPER(propertyvalue)='YES'
  )
  BEGIN
    -- If the note was opened from TEC, then we do not want to send a note back to TEC.
    -- This is determined by seeing if the note's subject matches the SystemConfiguration
    -- value TEC_EVENT_OWNERSHIP.TEC_NOTE_SUBJECT.
    SELECT @TEC_NOTE_SUBJECT = propertyvalue
    FROM SystemConfiguration WITH (NOLOCK)
    WHERE groupname='TEC_EVENT_OWNERSHIP'
      AND propertyname='TEC_NOTE_SUBJECT'

    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Looking for notes opened from TEC as determined by having Subject = ''%s''', 0, 1, @now, @TEC_NOTE_SUBJECT) WITH NOWAIT
    END
   
    INSERT INTO StagedTECStatus ( noteid, Status, ContextUserID )
      SELECT n.id, 'ACK', ISNULL(n.UserID,'')
      FROM @note AS n
      JOIN Note_V AS note WITH (NOLOCK) ON note.id = n.id
      WHERE note._Subject <> @TEC_NOTE_SUBJECT
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: error %d inserting into StagedTECStatus', 11, 1, @now, @err) WITH NOWAIT
      GOTO ErrorExit
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Inserted %d record(s) into StagedTECStatus', 0, 1, @now, @rc) WITH NOWAIT
    END
  END -- Handle notification to TEC
  
  -- If any notes are to be created in a Closed state, then update their status now
  IF EXISTS (SELECT 1 FROM @note WHERE FinalStatus = 'Closed')
  BEGIN
    UPDATE n SET _Status = 'Closed'
    FROM Note_A AS n
    JOIN @note AS nn ON nn.id = n.id
    WHERE nn.FinalStatus = 'Closed'
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: error %d closing notes', 3, 1, @now, @err) WITH NOWAIT
      GOTO ErrorExit
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: asisp_processStagedNotes: Closed %d newly created note(s)', 0, 1, @now, @rc) WITH NOWAIT
    END
  END -- If any notes are to be created in a Closed state, then update their status now

  -- Remove the processed records from StagedNote
  DELETE s
  FROM StagedNote AS s
  JOIN @note AS n ON n.id = s.id
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: error %d deleting from StagedNote', 11, 1, @now, @err) WITH NOWAIT
    GOTO ErrorExit
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Deleted %d record(s) from StagedNote', 0, 1, @now, @rc) WITH NOWAIT
  END
  
  IF @tcount = 0
    COMMIT TRANSACTION asisp_processStagedNotes

  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: asisp_processStagedNotes: Finished', 0, 1, @now) WITH NOWAIT
  END

  RETURN @rc

ErrorExit:
  IF @tcount IS NOT NULL
    ROLLBACK TRANSACTION asisp_processStagedNotes
  RETURN @rc
GO
