-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/etl_getash.sql
-- SID: 1.14
-- Date Modified: 05/01/04 11:08:43

-- Module description:
-- Create the etl_getash procedure which is used by
-- TDW to return alert state history data.

USE Object
IF EXISTS (SELECT * FROM sysobjects WHERE type = 'P' AND name = 'etl_getash')
  DROP PROCEDURE etl_getash
GO

-- File: archive/db/sql/procs/etl_getash.sql
-- SID: 1.14
-- Date Modified: 05/01/04 11:08:43

-- Module description:
-- Create the etl_getash procedure which is used by
-- TDW to return alert state history data.
-- The procedure also sets the mtime and status columns of the ash row of the twh_ExtractCtrl table.
-- The procedure also stores alert state history data in twh_defer_ash if the related resource does not
-- have a row in twh_Inventory

CREATE PROCEDURE etl_getash
  @verbose TINYINT=0
AS
  SET NOCOUNT ON

  DECLARE @now NVARCHAR(40)
  DECLARE @rc INT
  DECLARE @err INT
  DECLARE @sql NVARCHAR(2000)
  DECLARE @this_xid DECIMAL(18)
  DECLARE @last_xid DECIMAL(18)
  DECLARE @gmoffsetmin INT
  DECLARE @last_gmoffsetmin INT
  DECLARE @GMT_CHANGE_TIME DATETIME
  DECLARE @process_gmtc TINYINT

  CREATE TABLE #ash_set (
    ashid            INT NOT NULL,            -- The id associated with the alert state history. This will
                                              -- come from the ash_id column of the AlertStateHistory table.
    AlertStateID     INT NOT NULL,            -- The new TBSM alert state associated with the event: 1=Green, 
                                              -- 2=Yellow, 3=Red. This will come from the AlertStateID column 
                                              -- of the AlertStateHistory table
    OldAlertStateID  INT NOT NULL,            -- The old TBSM alert state associated with the event: 1=Green, 
                                              -- 2=Yellow, 3=Red. This will come from the OldAlertStateID column of the AlertStateHistory table
    ctime            DATETIME NOT NULL,       -- The time the alert state history record was created.
    event_type       NVARCHAR(32) NOT NULL,   -- `Alert State Change'
    evt_corr_val     NVARCHAR(40) NOT NULL,   -- `ASH\' + ashid
    twh_ash_id       INT NOT NULL,            -- 0
    sevt_corr_val    NVARCHAR(40) NOT NULL,   -- evt_cid + `\' + evt_id
    twh_sevt_id      INT NOT NULL,            -- 0
    evt_detail       NVARCHAR(255) NOT NULL,  -- from the AlertStateHistory table
    comp_id          DECIMAL(18,0)            -- TDW component id
  )

  IF @verbose >= 1 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getash starts', 0, 1, @now) WITH NOWAIT
  END

  IF @verbose >= 2 
  BEGIN
    SELECT 'twh_ExtractCtrl'='twh_ExtractCtrl',* FROM twh_ExtractCtrl WHERE name = 'ash'
  END

  IF EXISTS (SELECT 1 FROM twh_ExtractCtrl WHERE name = 'ash' AND status=0) 
  BEGIN
    -- if status=0, TDW told TBSM that the previous result set was successfully processed so we can
    -- delete any deferred rows we sent last time
    DELETE twh_defer_ash WHERE comp_id <> 0
    SELECT @rc = @@ROWCOUNT, @err = @@ERROR

    IF @err < 0 
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getash: error %d deleting sent rows from twh_defer_ash', 11, 1, @now, @err) WITH NOWAIT
      RETURN -30000
    END

    IF @verbose >= 1 
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getash deleted %d rows from twh_defer_ash that were sent to TDW last time', 0, 1, @now, @rc) WITH NOWAIT
    END
  END
  ELSE 
  BEGIN
    -- if status<>0, TDW did NOT tell TBSM that the previous result set was successfully processed.  We
    -- need to try again but we don't want to defer those rows that we deferred last time.  We'll start
    -- out by forgetting those we tried to send last time.
    DELETE twh_defer_ash WHERE age <= 1
    SELECT @rc = @@ROWCOUNT, @err = @@ERROR

    IF @err < 0 
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getash: error %d deleting last rows deferred from twh_defer_ash', 11, 1, @now, @err) WITH NOWAIT
      RETURN -30000
    END

    IF @verbose >= 1 
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getash deleted %d rows from twh_defer_ash that were deferred from last time', 0, 1, @now, @rc) WITH NOWAIT
    END
  END

  -- indicate that a result set has been prepared for TDW.
  UPDATE twh_ExtractCtrl SET status = 1 WHERE name = 'ash'

  -- add rows to twh_defer_ash first until we know their TDW comp_id
  SET @sql = 
  ' INSERT INTO twh_defer_ash ' +
  ' SELECT h.ash_id, ' +
  '        h.AlertStateID, ' +
  '        h.OldAlertStateID, ' +
  '        h.ctime, ' +
  '        ''Alert State Change'', ' +
  '        ''ASH\'' + CAST(h.ash_id AS VARCHAR(40)) AS evt_corr_val, ' +
  '        0 AS twh_ash_id, ' +
  '        h.evt_cid + ''\'' + CAST(h.evt_id AS VARCHAR(40)) AS sevt_corr_val, ' +
  '        0 AS twh_evt_id, ' +
  '        h.cid, ' +
  '        h.id, ' +
  '        0, ' +
  '        0, ' +
  '        h.evt_detail ' +
  '   FROM AlertStateHistory h WITH (NOWAIT)'

  -- What is the first event we'll return? If we're initializing use twh_PruneEventCtrl to control how far back we'll go.
  IF EXISTS (SELECT 1 FROM twh_ExtractCtrl WHERE name = 'ash' AND last_xid = 0) 
  BEGIN
    SET @sql = @sql + ' WHERE h.cid = ''LOB'' AND DATEDIFF(dd, h.ctime, CURRENT_TIMESTAMP) <= (SELECT event_age FROM twh_PruneEventCtrl)'
  END
  ELSE 
  BEGIN
    SELECT @this_xid = this_xid, @last_xid = last_xid FROM twh_ExtractCtrl WHERE name = 'ash'
    SET @sql = @sql +
      ' WHERE h.cid = ''LOB'' AND ash_id >  ' + CAST(@last_xid AS NVARCHAR(18)) +
      '   AND ash_id <= ' + CAST(@this_xid AS NVARCHAR(18))
  END

  IF @verbose >= 2 
  BEGIN
    PRINT @sql
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getash executing: %s', 0, 1, @now, @sql) WITH NOWAIT
  END
  EXEC(@sql)
  SELECT @rc=@@ROWCOUNT, @err=@@ERROR

  IF @err < 0 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getash: error %d inserting into twh_defer_ash', 11, 1, @now, @err) WITH NOWAIT
    RETURN -30000
  END

  IF @verbose >= 1 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getash inserted %d rows into twh_defer_ash', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- backfill TDW comp_id's - this may hit rows we just added or rows that were deferred from
  -- previous ETLs
  UPDATE d SET comp_id = i.comp_id
    FROM twh_defer_ash d
    JOIN twh_Inventory i ON i.cid = d.cid AND i.id = d.id
  SELECT @rc = @@ROWCOUNT, @err = @@ERROR

  IF @err < 0 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getash: error %d setting comp_id''s twh_defer_ash', 11, 1, @now, @err) WITH NOWAIT
    RETURN -30000
  END

  IF @verbose >= 1 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getash updated comp_id of %d rows of twh_defer_ash', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- adjust timestamps in result set to be GMT-based

  EXEC etl_getgmtvars 'ash', @process_gmtc OUTPUT, @gmoffsetmin OUTPUT, @last_gmoffsetmin OUTPUT, @GMT_CHANGE_TIME OUTPUT, @verbose=@verbose

  IF @process_gmtc = 0 -- adjust by a single offset
    UPDATE twh_defer_ash SET ctime = DATEADD(MI, @gmoffsetmin, ctime) WHERE age = 0
  ELSE                 -- adjust by the two offsets
    UPDATE twh_defer_ash SET ctime = CASE
                                       WHEN ctime >= @GMT_CHANGE_TIME THEN DATEADD(MI, @gmoffsetmin, ctime)
                                       ELSE                                DATEADD(MI, @last_gmoffsetmin, ctime)
                                     END
           WHERE age = 0

  -- Pull out the rows that have TDW comp_id's - these are what we'll return in the result set
  INSERT INTO #ash_set
    SELECT
      ashid,
      AlertStateID,
      OldAlertStateID,
      ctime,
      event_type,
      evt_corr_val,
      twh_ash_id,
      sevt_corr_val,
      twh_sevt_id,
      evt_detail,
      comp_id
    FROM twh_defer_ash
    WHERE comp_id <> 0
  SELECT @rc = @@ROWCOUNT, @err = @@ERROR

  IF @err < 0 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getash: error %d inserting into #ash_set', 11, 1, @now, @err) WITH NOWAIT
    RETURN -30000
  END

  IF @verbose >= 1 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getash inserted %d rows into temporary result set table', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- increment the age of the deferred rows so we know which were added
  -- this time in case TDW does not process this result set and we need
  -- to start over with the same beginning extraction id (last_xid)
  UPDATE twh_defer_ash SET age = age + 1

  SELECT * FROM #ash_set

  IF @verbose >= 1 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getash returning', 0, 1, @now) WITH NOWAIT
  END
GO
