-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/etl_getdeletedobjs.sql
-- SID: 1.10
-- Date Modified: 05/01/07 10:34:45

-- Module description:
-- Create the etl_getupdateattr procedure which is used by TDW to return a result set of objects that
-- have been physically deleted (appear in twh_Inventory but are no longer physically present in the TBSM database

USE Object
IF EXISTS (SELECT * FROM sysobjects WHERE type = 'P' AND name = 'etl_getdeletedobjs')
  DROP PROCEDURE etl_getdeletedobjs
GO
-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}

-- File: archive/db/sql/procs/etl_getdeletedobjs.sql
-- SID: 1.10
-- Date Modified: 05/01/07 10:34:45

-- Module description:
-- Create the etl_getupdateattr procedure which is used by TDW to return a result set of objects that
-- have been physically deleted (appear in twh_Inventory but are no longer physically present in the TBSM database
-- The result set has the following format
--    cid         ClassID NOT NULL
--    id          ObjID   NOT NULL
--    comp_id     DECIMAL(20,0)  NOT NULL
--    mtime       DATETIME NOT NULL
--    attr_value  NVARCHAR(256) NOT NULL
-- The procedure also updates the mtime and status columns of the twh_ResourceDelete row of the twh_ExtractCtrl table

CREATE PROCEDURE etl_getdeletedobjs
  @verbose TINYINT=0
AS
  SET NOCOUNT ON
  DECLARE @now NVARCHAR(40)
  DECLARE @err      INTEGER
  DECLARE @rc       INTEGER
  DECLARE @last_xid DECIMAL
  DECLARE @this_xid DECIMAL
  DECLARE @gmoffsetmin INT
  DECLARE @last_gmoffsetmin INT
  DECLARE @GMT_CHANGE_TIME DATETIME
  DECLARE @process_gmtc TINYINT

  CREATE TABLE #resultSet (
    cid         ClassID NOT NULL,
    id          ObjID   NOT NULL,
    comp_id     DECIMAL(20,0)  NOT NULL,
    mtime       DATETIME NOT NULL
  )

  IF @verbose >= 1 BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getdeletedobjs starts', 0, 1, @now) WITH NOWAIT
  END

  -- standard transaction logic
  DECLARE @tcount INT
  SELECT @tcount = @@TRANCOUNT
  IF @tcount = 0
    BEGIN TRAN etl_getdeletedobjs
  ELSE
    SAVE TRAN etl_getdeletedobjs

  UPDATE twh_ExtractCtrl WITH (HOLDLOCK) SET this_mtime = CURRENT_TIMESTAMP, status = 1
  WHERE xtable = 'twh_ResourceDelete'

  SELECT @last_xid = last_xid, @this_xid = this_xid 
  FROM twh_ExtractCtrl
  WHERE xtable = 'twh_ResourceDelete'

  INSERT INTO #resultSet (cid,   id,   comp_id, mtime)
  SELECT                  d.cid, d.id, 0,       d.ctime
  FROM twh_ResourceDelete AS d WITH (NOLOCK)
  WHERE d.xid > @last_xid 
    AND d.xid <= @this_xid
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err < 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getdeletedobjs: error %d inserting into #resultSet', 11, 1, @now, @err) WITH NOWAIT
    ROLLBACK TRANSACTION etl_getdeletedobjs
    RETURN -30000
  END

  UPDATE r SET comp_id = i.comp_id
  FROM #resultSet as r
    JOIN twh_Inventory AS i ON i.cid = r.cid AND i.id = r.id
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err < 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getdeletedobjs: error %d updating #resultSet', 11, 1, @now, @err) WITH NOWAIT
    ROLLBACK TRANSACTION etl_getdeletedobjs
    RETURN -30000
  END

  -- adjust timestamps in result set to be GMT-based

  EXEC etl_getgmtvars 'deletedobjs', @process_gmtc OUTPUT, @gmoffsetmin OUTPUT, @last_gmoffsetmin OUTPUT, @GMT_CHANGE_TIME OUTPUT, @verbose=@verbose

  IF @process_gmtc = 0 -- adjust by a single offset
    UPDATE #resultSet SET mtime = DATEADD(MI, @gmoffsetmin, mtime)
  ELSE                 -- adjust by the two offsets
    UPDATE #resultSet SET mtime = CASE
                                       WHEN mtime >= @GMT_CHANGE_TIME THEN DATEADD(MI, @gmoffsetmin, mtime)
                                       ELSE                                DATEADD(MI, @last_gmoffsetmin, mtime)
                                     END

  SELECT * FROM #resultSet

  -- Finish up TRANSACTION
  IF @tcount = 0
    COMMIT TRAN etl_getdeletedobjs


  IF @verbose >= 1 BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getdeletedobjs returning', 0, 1, @now) WITH NOWAIT
  END
GO