-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/etl_getevents.sql
-- SID: 1.20
-- Date Modified: 05/01/07 10:34:48

-- Module description:
-- Create the etl_getevents procedure

USE Object
IF EXISTS (SELECT * FROM sysobjects WHERE type = 'P' AND name = 'etl_getevents')
  DROP PROCEDURE etl_getevents
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/etl_getevents.sql
-- SID: 1.20
-- Date Modified: 05/01/07 10:34:48

-- Module description:
-- Create the etl_getevents procedure which is used by
-- TDW to generate a result set of events (messages,
-- exceptions, and child events).  It returns a result
-- set with the following columns:
--    obj_cid          ClassID NOT NULL,        -- Class id of the resource the event belongs to
--    obj_id           ObjID NOT NULL,          -- Object id of the resource the event belongs to
--    comp_id          DECIMAL(20,0) NOT NULL,  -- The TDW component id associated with the resource the event belongs to. 
--                                              -- This will come from the comp_id column of the twh_Inventory table, based
--                                              -- on the cid and id of the resource the event belongs to.
--                                              -- The value will never be zero.  If we can't find the row in the inventory table,
--                                              -- we'll save the row from the result set in the tdw_defer_events table.
--                                              -- Such rows will be included in the result set the next time etl_getevents is called 
--                                              -- and the comp_id can be found in the inventory table.
--    cid              ClassID NOT NULL,        -- Class id of the event (`MESG', `EXCP', or `CHEV')
--    id               ObjID NOT NULL,          -- Object id of the event
--    event_type       ObjName NOT NULL,        -- This will come from the twh column of the twh_tbsm_map table, based on the event cid
--    name             NVARCHAR(255) NOT NULL,  -- Name associated with the event
--    ctime            DATETIME NOT NULL,       -- The time the event was created
--    detail           NVARCHAR(255) NULL,      -- 
--    alertstate       INT NOT NULL,            -- The TBSM alert state associated with the event: 1=Green, 2=Yellow, 3=Red.
--    priority         INT NOT NULL,            -- The TBSM priority associated with the event: 1=Critical, 2=High, 3=Medium, 4=Low, 5=Ignore.
--    twh_cid_id_val   NVARCHAR(40) NULL,       -- cid + `\' + CONVERT(NVARCHAR(40), evt_id) or NULL if evt_id = 0
--    twh_evt_corr_val NVARCHAR(40),            -- `gtmevt\' + CONVERT(NVARCHAR(40), id)
--    twh_evt_id       INT NOT NULL DEFAULT(0)
-- The procedure sets the mtime and status columns of the events row of the twh_ExtractCtrl
-- table.
-- The procedure also saves event data in twh_defer_events if the related resource does not have a row
-- in twh_Inventory yet

CREATE PROCEDURE etl_getevents
  @verbose TINYINT=0
AS
  SET NOCOUNT ON
  DECLARE @now NVARCHAR(40)
  DECLARE @rc INT
  DECLARE @err INT
  DECLARE @sql NVARCHAR(2000)
  DECLARE @this_xid DECIMAL(18)
  DECLARE @last_xid DECIMAL(18)
  DECLARE @gmoffsetmin INT
  DECLARE @currgmdate DATETIME
  DECLARE @last_gmoffsetmin INT
  DECLARE @lastgmdate DATETIME
  DECLARE @GMT_CHANGE_TIME DATETIME
  DECLARE @process_gmtc TINYINT

  CREATE TABLE #event_set (
    obj_cid          ClassID NOT NULL,        -- Class id of the resource the event belongs to
    obj_id           ObjID NOT NULL,          -- Object id of the resource the event belongs to
    comp_id          DECIMAL(20,0) NOT NULL,  -- The TDW component id associated with the resource the event belongs to. 
                                              -- This will come from the comp_id column of the twh_Inventory table, based
                                              -- on the cid and id of the resource the event belongs to.
                                              -- The value will never be zero.  If we can't find the row in the inventory table,
                                              -- we'll save the row from the result set in the tdw_defer_events table.
                                              -- Such rows will be included in the result set the next time etl_getevents is called 
                                              -- and the comp_id can be found in the inventory table.
    cid              ClassID NOT NULL,        -- Class id of the event (`MESG', `EXCP', or `CHEV')
    id               ObjID NOT NULL,          -- Object id of the event
    event_type       ObjName NOT NULL,        -- This will come from the twh column of the twh_tbsm_map table, based on the event cid
    name             NVARCHAR(255) NOT NULL,  -- Name associated with the event
    ctime            DATETIME NOT NULL,       -- The time the event was created
    detail           NVARCHAR(255) NULL,      -- 
    alertstate       INT NOT NULL,            -- The TBSM alert state associated with the event: 1=Green, 2=Yellow, 3=Red.
    priority         INT NOT NULL,            -- The TBSM priority associated with the event: 1=Critical, 2=High, 3=Medium, 4=Low, 5=Ignore.
    twh_cid_id_val   NVARCHAR(40) NULL,       -- cid + `\' + CONVERT(NVARCHAR(40), evt_id) or NULL if evt_id = 0
    twh_evt_corr_val NVARCHAR(40),            -- `gtmevt\' + CONVERT(NVARCHAR(40), id)
    twh_evt_id       INT NOT NULL DEFAULT(0)
  )

  IF @verbose >= 1 
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getevents starts', 0, 1, @now) WITH NOWAIT
  END

  IF @verbose >= 2 
  BEGIN
    SELECT 'twh_ExtractCtrl'='twh_ExtractCtrl',* FROM twh_ExtractCtrl WHERE name = 'events'
  END

  IF EXISTS (SELECT 1 FROM twh_ExtractCtrl WHERE name = 'events' AND status=0) 
  BEGIN
    -- if status=0, TDW told TBSM that the previous result set was successfully processed so we can
    -- delete any deferred rows we sent last time
    DELETE twh_defer_events WHERE comp_id <> 0
    SELECT @rc=@@ROWCOUNT, @err=@@ERROR

    IF @err < 0 
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getevents: error %d deleting sent rows from twh_defer_events', 11, 1, @now, @rc) WITH NOWAIT
      RETURN -30000
    END

    IF @verbose >= 1 
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getevents deleted %d rows from twh_defer_events that were sent to TDW last time', 0, 1, @now, @rc) WITH NOWAIT
    END
  END
  ELSE 
  BEGIN
    -- if status<>0, TDW did NOT tell TBSM that the previous result set was successfully processed.  We
    -- need to try again but we don't want to defer those events that we deferred last time.  We'll start
    -- out by forgetting those we tried to send last time.
    DELETE twh_defer_events WHERE age <= 1
    SELECT @rc=@@ROWCOUNT, @err=@@ERROR

    IF @err < 0 
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getevents: error %d deleting last rows added from twh_defer_events', 11, 1, @now, @rc) WITH NOWAIT
      RETURN -30000
    END

    IF @verbose >= 1 
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getevents deleted %d rows from twh_defer_events that were deferred from last time', 0, 1, @now, @rc) WITH NOWAIT
    END
  END

  -- indicate that a result set has been prepared for TDW.
  UPDATE twh_ExtractCtrl SET status = 1 WHERE name = 'events'

  ---------------------------------------------------------------------------------------------------
  -- Begin Init
  -- add rows to twh_deffered_events first for the event in Message_V, Exception_V, and ChildEvent_V
  -- This is only used for the first time initial load because new notes may depend on these events. 
  ----------------------------------------------------------------------------------------------------
  -- for Message_V
  IF EXISTS (SELECT 1 FROM twh_ExtractCtrl WHERE name = 'init_mesg' AND status = 1)
  BEGIN 
    SET @sql = 
    ' INSERT INTO twh_defer_events ' +
    ' SELECT e._ParentCID,  ' +
    '     e._ParentID,  ' +
    '     0, ' + -- use a 0 for comp_id for now, we''ll back-fill what we can later and defer those that can''t be resolved
    '     ''MESG'', ' +
    '     e.id, ' +
    '     m.twh, ' +
    '     e.name, ' +
    '     e.ctime, ' +
    '     e._REASON_CD, ' +
    '     e._AlertState, ' +
    '     e._Priority, ' +
    '     ''MESG\'' + CAST(e.id AS NVARCHAR(40)), ' +
    '     ''gtmmesg\'' + CAST(e.id AS NVARCHAR(40)), ' +
    '     0, ' +
    '     0 ' +
    ' FROM Message_V        AS e WITH (NOLOCK), ' +
    ' twh_tbsm_map AS m, ' +
    ' twh_ExtractCtrl c  ' +
    ' WHERE m.tbsm = ''MESG'' ' +
    ' AND m.map_type = ''EventType'' ' +
    ' AND c.this_xid >= e.id ' +
    ' AND c.name = ''init_mesg'''
    EXEC(@sql)
    SELECT @rc=@@ROWCOUNT, @err=@@ERROR

    IF @err < 0 BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getevents: error %d inserting from Message_V into twh_defer_events', 11, 1, @now, @rc) WITH NOWAIT
      RETURN -30000
    END

    IF @verbose >= 1 BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getevents inserted %d Message_V rows into twh_defer_events', 0, 1, @now, @rc) WITH NOWAIT
    END

  END

  -- for Exception_V
  IF EXISTS (SELECT 1 FROM twh_ExtractCtrl WHERE name = 'init_excp' AND status = 1)
  BEGIN
    SET @sql = NULL
    SET @sql = 
    ' INSERT INTO twh_defer_events ' +
    ' SELECT e._ParentCID,  ' +
    '     e._ParentID,  ' +
    '     0, ' + -- use a 0 for comp_id for now, we''ll back-fill what we can later and defer those that can''t be resolved
    '     ''EXCP'', ' +
    '     e.id, ' +
    '     m.twh, ' +
    '     e.name, ' +
    '     e.ctime, ' +
    '     e._EXCP_CD, ' +
    '     e._AlertState, ' +
    '     e._Priority, ' +
    '     ''EXCP\'' + CAST(e.id AS NVARCHAR(40)), ' +
    '     ''gtmexcp\'' + CAST(e.id AS NVARCHAR(40)), ' +
    '     0, ' +
    '     0 ' +
    ' FROM Exception_V        AS e WITH (NOLOCK), ' +
    ' twh_tbsm_map AS m, ' +
    ' twh_ExtractCtrl c  ' +
    ' WHERE m.tbsm = ''EXCP'' ' +
    ' AND m.map_type = ''EventType'' ' +
    ' AND c.this_xid >= e.id ' +
    ' AND c.name = ''init_excp'''
    EXEC(@sql)
    SELECT @rc=@@ROWCOUNT, @err=@@ERROR

    IF @err < 0 BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getevents: error %d inserting from Exception_V into twh_defer_events', 11, 1, @now, @rc) WITH NOWAIT
      RETURN -30000
    END

    IF @verbose >= 1 BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getevents inserted %d Exception_V rows into twh_defer_events', 0, 1, @now, @rc) WITH NOWAIT
    END
  END

  -- for ChildEvent_V
  IF EXISTS (SELECT 1 FROM twh_ExtractCtrl WHERE name = 'init_chev' AND status = 1)
  BEGIN
    SET @sql =
    ' INSERT INTO twh_defer_events ' +
    ' SELECT e._ParentCID,  ' +
    '     e._ParentID,  ' +
    '     0, ' + -- use a 0 for comp_id for now, we''ll back-fill what we can later and defer those that can''t be resolved
    '     ''CHEV'', ' +
    '     e.id, ' +
    '     m.twh, ' +
    '     e.name, ' +
    '     e.ctime, ' +
    '     e._EventDetail, ' +
    '     e._AlertState, ' +
    '     e._Priority, ' +
    '     ''CHEV\'' + CAST(e.id AS NVARCHAR(40)),  ' +
    '     ''gtmchev\'' + CAST(e.id AS NVARCHAR(40)), ' +
    '     0, ' +
    '     0 ' +
    ' FROM ChildEvent_V        AS e WITH (NOLOCK), ' +
    ' twh_tbsm_map AS m, ' +
    ' twh_ExtractCtrl c  ' +
    ' WHERE m.tbsm = ''CHEV'' ' +
    ' AND m.map_type = ''EventType'' ' +
    ' AND c.this_xid >= e.id ' +
    ' AND c.name = ''init_chev'''
    EXEC(@sql)
    SELECT @rc=@@ROWCOUNT, @err=@@ERROR

    IF @err < 0 BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getevents: error %d inserting from ChildEvent_V into twh_defer_events', 11, 1, @now, @rc) WITH NOWAIT
      RETURN -30000
    END

    IF @verbose >= 1 BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: etl_getevents inserted %d ChildEvent_V rows into twh_defer_events', 0, 1, @now, @rc) WITH NOWAIT
    END
  END
  --------------------------------------------------------------------------------------------
  -- END Init
  --------------------------------------------------------------------------------------------

  -- add rows to twh_deffered_events first until we know their TDW comp_id
  SET @sql =
  ' INSERT INTO twh_defer_events ' +
  ' SELECT e.obj_cid,  ' +
  '        e.obj_id,  ' +
  '        0, ' + -- use a 0 for comp_id for now, we''ll back-fill what we can later and defer those that can''t be resolved
  '        e.cid, ' +
  '        e.evt_id, ' +
  '        m.twh, ' +
  '        e.name, ' +
  '        e.ctime, ' +
  '        e.detail, ' +
  '        e.alertstate, ' +
  '        e.priority, ' +
  '        e.cid + ''\'' + CAST(e.evt_id AS NVARCHAR(40)),  ' +
  '        ''gtmevt\'' + CAST(e.id AS NVARCHAR(40)), ' +
  '        0, ' +
  '        0 ' +
  '   FROM event        AS e WITH (NOLOCK) ' +
  '   JOIN twh_tbsm_map AS m ON e.cid = m.tbsm ' +
  '   WHERE m.map_type = ''EventType'''

  -- What is the first event we'll return? If we're initializing use twh_PruneEventCtrl to control how far back we'll go.
  IF EXISTS (SELECT 1 FROM twh_ExtractCtrl WHERE name = 'events' AND last_xid = 0) BEGIN
    SET @sql = @sql + '
      AND DATEDIFF(dd, e.ctime, CURRENT_TIMESTAMP) <= (SELECT event_age FROM twh_PruneEventCtrl)'
  END
  ELSE BEGIN
    SELECT @this_xid = this_xid, @last_xid = last_xid FROM twh_ExtractCtrl WHERE name = 'events'
    SET @sql = @sql + '
      AND CAST(CAST(tstamp AS BIGINT) AS DECIMAL(18)) >  ' + CAST(@last_xid AS NVARCHAR(18)) + '
      AND CAST(CAST(tstamp AS BIGINT) AS DECIMAL(18)) <= ' + CAST(@this_xid AS NVARCHAR(18))
  END

  IF @verbose >= 2 BEGIN
    PRINT @sql
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getevents executing: %s', 0, 1, @now, @sql) WITH NOWAIT
  END
  EXEC(@sql)
  SELECT @rc=@@ROWCOUNT, @err=@@ERROR

  IF @err < 0 BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getevents: error %d inserting from event into twh_defer_events', 11, 1, @now, @rc) WITH NOWAIT
    RETURN -30000
  END

  IF @verbose >= 1 BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getevents inserted %d event rows into twh_defer_events', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- backfill TDW comp_id's - this may hit rows we just added or rows that were deferred from
  -- previous ETLs
  UPDATE d SET comp_id = i.comp_id
    FROM twh_defer_events d
    JOIN twh_Inventory       i ON i.cid = d.obj_cid AND i.id = d.obj_id
  SELECT @rc = @@ROWCOUNT, @err = @@ERROR

  IF @err < 0 BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getevents: error %d backfilling comp_id''s', 11, 1, @now, @err) WITH NOWAIT
    RETURN -30000
  END

  IF @verbose >= 1 BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getevents updated comp_id of %d rows of twh_defer_events', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- adjust timestamps in result set to be GMT-based

  EXEC etl_getgmtvars 'events', @process_gmtc OUTPUT, @gmoffsetmin OUTPUT, @last_gmoffsetmin OUTPUT, @GMT_CHANGE_TIME OUTPUT

  IF @process_gmtc = 0 -- adjust by a single offset
    UPDATE twh_defer_events SET ctime = DATEADD(MI, @gmoffsetmin, ctime) WHERE age = 0
  ELSE                 -- adjust by the two offsets
    UPDATE twh_defer_events SET ctime = CASE
                                          WHEN ctime >= @GMT_CHANGE_TIME THEN DATEADD(MI, @gmoffsetmin, ctime)
                                          ELSE                                DATEADD(MI, @last_gmoffsetmin, ctime)
                                        END
           WHERE age = 0

  -- Pull out the rows that have TDW comp_id's - these are what we'll return in the result set
  INSERT INTO #event_set
    SELECT
      obj_cid,
      obj_id,
      comp_id,
      cid,
      id,
      event_type,
      name,
      ctime,
      detail,
      alertstate,
      priority,
      twh_cid_id_val,
      twh_evt_corr_val,
      twh_evt_id
    FROM twh_defer_events
    WHERE comp_id <> 0
  SELECT @rc = @@ROWCOUNT, @err = @@ERROR

  IF @err < 0 BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getevents: error %d inserting into #event_set', 11, 1, @now, @err) WITH NOWAIT
    RETURN -30000
  END

  IF @verbose >= 1 BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getevents inserted %d rows into temporary result set table', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- increment the age of the deferred events so we know which were added
  -- this time in case TDW does not process this result set and we need
  -- to start over with the same beginning extraction id (last_xid)
  UPDATE twh_defer_events SET age = age + 1

  SELECT * FROM #event_set

  IF @verbose >= 1 BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getevents returning', 0, 1, @now) WITH NOWAIT
  END
GO
