-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/etl_getgmtvars.sql
-- SID: 1.5
-- Date Modified: 04/11/14 11:55:16

-- Module description:
-- Create the etl_getgmtvars procedure

USE Object
IF EXISTS (SELECT * FROM sysobjects WHERE type = 'P' AND name = 'etl_getgmtvars')
  DROP PROCEDURE etl_getgmtvars
GO

-- File: archive/db/sql/procs/etl_getgmtvars.sql
-- SID: 1.5
-- Date Modified: 04/11/14 11:55:16

-- Module description:
-- Create the etl_getgmtvars procedure which is used by
-- other procedures to help convert timestamps expressed
-- in local time to GMT-based timestamps.

-- The basic idea is that every timestamp should get adjusted by a
-- certain number of minutes.  In many instances, all timestamps
-- will get adjusted by the same number of minutes.  Sometimes,
-- timestamps will be adjusted by one of two values.  This happens
-- you switch from daylight savings time to standard time or vice
-- versa: timestamps prior to the time change are adjusted by one
-- amount and timestamps after the time change are adjusted by
-- another amount.

-- Callers are expected to call this procedure in the following
-- manner:
--
--  DECLARE @gmoffsetmin INT
--  DECLARE @last_gmoffsetmin INT
--  DECLARE @GMT_CHANGE_TIME DATETIME
--  DECLARE @process_gmtc TINYINT
--  .
--  .
--  .
--  EXEC etl_getgmtvars @loadtypestring, @process_gmtc OUTPUT, @gmoffsetmin OUTPUT, @last_gmoffsetmin OUTPUT, @GMT_CHANGE_TIME OUTPUT
--
--  IF @process_gmtc = 0 -- adjust by a single offset
--    UPDATE mytable SET mytimestamp = DATEADD(MI, @gmoffsetmin, mytimestamp)
--  ELSE BEGIN -- adjust by the two offsets
--    UPDATE mytable SET mytimestamp = CASE
--                                       WHEN mytimestamp >= @GMT_CHANGE_TIME THEN DATEADD(MI, @gmoffsetmin, mytimestamp)
--                                       ELSE                                      DATEADD(MI, @last_gmoffsetmin, mytimestamp)
--                                     END
--  END

CREATE PROCEDURE etl_getgmtvars
  @loadtype NVARCHAR(50),            -- a unique string used as an index to the twh_gmt_time table to record previous and current information
  @process_gmtc TINYINT OUTPUT,      -- a value of 0 indicates tha only one offset is needed for adjustments
                                     -- a value of 1 indicates that two offsets are needed for adjustments
  @gmoffsetmin INT OUTPUT,           -- the current offset from GMT in minutes (west of GMT is positive, east is negative)
  @last_gmoffsetmin INT OUTPUT,      -- the previous offset from GMT in minutes (west of GMT is positive, east is negative) - this may be NULL if
                                     -- a previous offset was not recorded
  @GMT_CHANGE_TIME DATETIME OUTPUT,  -- the local timestamp at which the next time change takes place - for example: 2004/10/31 02:00:00.000
  @verbose TINYINT=0                 -- a flag that will enable diagnostic information when the procedure is invoked
AS
  SET NOCOUNT ON

  DECLARE @lastgmdate DATETIME
  DECLARE @now DATETIME
  DECLARE @gmt DATETIME
  DECLARE @nowstring NVARCHAR(40)

  -- figure out current GMT offset
  SELECT @now = CURRENT_TIMESTAMP, @gmt = GETUTCDATE()
  SET @gmoffsetmin = DATEDIFF(mi, @now, @gmt)

  -- get previous GMT offset
  SELECT @last_gmoffsetmin = last_gmt, @lastgmdate = last_date FROM twh_gmt_time WHERE loadtype = @loadtype

  -- save current GMT offset
  UPDATE twh_gmt_time SET this_gmt = @gmoffsetmin, this_date = @now WHERE loadtype = @loadtype

  -- assume we don't have only one GMT offset to worry about - don't worry about daylight/standard changes and we'll
  -- just adjust by the current GMT offset
  SET @process_gmtc = 0

  -- Are the offsets different
  IF @last_gmoffsetmin IS NOT NULL AND @gmoffsetmin <> @last_gmoffsetmin
  BEGIN
    SELECT @GMT_CHANGE_TIME = propertyvalue 
      FROM SystemConfiguration 
      WHERE propertyname = 'GMT_CHANGE_TIME'
      AND groupname = 'TWH'

    -- We'll worry about two offsets if and only if:
    --   The date that the last offset was acquired was prior to the date at which the offset changes (GMT_CHANGE_TIME)
    --   The current date is after the date at which the offset changes (GMT_CHANGE_TIME)
    IF (DATEDIFF(MI, @lastgmdate, @GMT_CHANGE_TIME) > 0) AND (DATEDIFF(MI, @GMT_CHANGE_TIME, @now) > 0) 
      SELECT @process_gmtc = 1 
  END

  IF @verbose >= 1
  BEGIN
    SET @nowstring = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getgmtvars: process_gmtc=%d, gmoffsetmin=%d, last_gmoffsetmin=%d', 0, 1, @nowstring, @process_gmtc, @gmoffsetmin, @last_gmoffsetmin) WITH NOWAIT
  END

  RETURN
GO