-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/etl_getupdateattr.sql
-- SID: 1.13
-- Date Modified: 05/01/07 10:34:50

-- Module description:
-- Create the etl_getupdateattr procedure

USE Object
IF EXISTS (SELECT * FROM sysobjects WHERE type = 'P' AND name = 'etl_getupdateattr')
  DROP PROCEDURE etl_getupdateattr
GO
-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}

-- File: archive/db/sql/procs/etl_getupdateattr.sql
-- SID: 1.13
-- Date Modified: 05/01/07 10:34:50

-- Module description:
-- Create the etl_getupdateattr procedure which is used by TDW to return a result set of attribute changes
-- The result set has the following format
--    cid         ClassID NOT NULL
--    id          ObjID   NOT NULL
--    comp_id     DECIMAL(20,0)  NOT NULL
--    attr_type   NVARCHAR(256) NOT NULL
--    attr_value  NVARCHAR(256) NOT NULL
-- The procedure also updates the twh_AttrUpdate row of twh_ExtractCtrl: mtime is set to the current time
-- and status is changed to 1

CREATE PROCEDURE etl_getupdateattr
  @verbose TINYINT=0
AS
  SET NOCOUNT ON
  DECLARE @now NVARCHAR(40)
  DECLARE @last_xid DECIMAL
  DECLARE @this_xid DECIMAL
  DECLARE @err      INTEGER
  DECLARE @rc       INTEGER
  DECLARE @gmoffsetmin INT
  DECLARE @currgmdate DATETIME
  DECLARE @last_gmoffsetmin INT
  DECLARE @lastgmdate DATETIME
  DECLARE @GMT_CHANGE_TIME DATETIME
  DECLARE @process_gmtc TINYINT

  CREATE TABLE #resultSet (
    cid         ClassID NOT NULL,
    id          ObjID   NOT NULL,
    comp_id     DECIMAL(20,0)  NOT NULL,
    attr_type   NVARCHAR(256) NOT NULL,
    attr_value  NVARCHAR(256) NOT NULL,
    mtime       DATETIME,
    comp_corr_val NVARCHAR(254) NOT NULL
  )

  IF @verbose >= 1 BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getupdateattr starts', 0, 1, @now) WITH NOWAIT
  END

  -- standard transaction logic
  DECLARE @tcount INT
  SELECT @tcount = @@TRANCOUNT
  IF @tcount = 0
    BEGIN TRAN etl_getupdateattr
  ELSE
    SAVE TRAN etl_getupdateattr

  UPDATE twh_ExtractCtrl SET this_mtime = CURRENT_TIMESTAMP, status = 1
  WHERE xtable = 'twh_AttrUpdate'
  
  SELECT @last_xid = last_xid, @this_xid = this_xid 
  FROM twh_ExtractCtrl
  WHERE xtable = 'twh_AttrUpdate'

  INSERT INTO #resultSet (cid,   id,   comp_id, attr_type,   attr_value, mtime, comp_corr_val)
  SELECT a.cid, 
        a.id, 
        0,       
        CASE a.attr_name WHEN 'deleted' THEN 'GTM_DELETED' 
        WHEN 'name' THEN 
          CASE m._public 
            WHEN 0 THEN 'NAME' 
            ELSE 'USER_LABEL'
          END
        ELSE a.attr_name END,  
        a.attr_value, 
        ctime,
        RTRIM(a.cid) + '|' + CONVERT(NVARCHAR, a.id)
  FROM twh_AttrUpdate  AS a WITH (NOLOCK)
    JOIN twh_class_map AS m ON m.cid = a.cid
    JOIN (SELECT cid,id,attr_name,max(xid) AS maxid FROM twh_AttrUpdate GROUP BY cid, id, attr_name) AS u
      ON u.cid = a.cid AND u.id = a.id AND u.attr_name = a.attr_name and u.maxid = a.xid
  WHERE a.xid > @last_xid 
    AND a.xid <= @this_xid
    AND a.attr_name <> 'description' -- 143279: do not report description changes
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err < 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getupdateattr: error %d inserting into #resultSet', 11, 1, @now, @err) WITH NOWAIT
    ROLLBACK TRANSACTION etl_getupdateattr
    RETURN -30000
  END

  IF @verbose >= 1 BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getupdateattr added %d rows to #resultSet', 0, 1, @now, @rc) WITH NOWAIT
  END

  UPDATE #resultSet SET comp_id = i.comp_id
  FROM #resultSet as r
    JOIN twh_Inventory AS i ON i.cid = r.cid AND i.id = r.id
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err < 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getupdateattr: error %d updating #resultSet', 11, 1, @now, @err) WITH NOWAIT
    ROLLBACK TRANSACTION etl_getupdateattr
    RETURN -30000
  END

  IF @verbose >= 1 BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getupdateattr updated comp_id''s for %d rows in #resultSet', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- adjust timestamps in result set to be GMT-based

  EXEC etl_getgmtvars 'updateattr', @process_gmtc OUTPUT, @gmoffsetmin OUTPUT, @last_gmoffsetmin OUTPUT, @GMT_CHANGE_TIME OUTPUT

  IF @process_gmtc = 0 -- adjust by a single offset
    UPDATE #resultSet SET mtime = DATEADD(MI, @gmoffsetmin, mtime)
  ELSE                 -- adjust by the two offsets
    UPDATE #resultSet SET mtime = CASE
                                    WHEN mtime >= @GMT_CHANGE_TIME THEN DATEADD(MI, @gmoffsetmin, mtime)
                                    ELSE                                DATEADD(MI, @last_gmoffsetmin, mtime)
                                  END

  SELECT * FROM #resultSet

  IF @verbose >= 1 BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: etl_getupdateattr returning', 0, 1, @now) WITH NOWAIT
  END

  -- Finish up TRANSACTION
  IF @tcount = 0
    COMMIT TRAN etl_getupdateattr

GO
